#This file consists of a suite of methods that interface with Regina triangulations of 3-manifolds and allow
#encoding and decoding of oriented isomorphism signatures, Pachner arc encodings, and directed Pachner path encodings.





import regina
import math

#A class that handles how to encode integers as printable characters. In particular this class is used in encodings of 
# isomorphism signatures, oriented isomorphism signatures, Pachner arc encodings, and directed Pachner path encodings.

class Pi:
    def __init__(self, d):          # d is the number of characters needed to encode the number of tets and terms in the destination sequnece 
        self.d = d

    # list to encode integers as printable characters based on their index in the list
    list = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',         
            'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
            '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '-']

    
    def decode(code):                    # function which takes in a string and outputs its integer value according to Pi.list 
        num = 0
        for i in range(len(code)):
            num += Pi.list.index(code[i]) * (64**i)
        return num 
    
    def encode(self, num):              # function which takes in an integer and outputs a string encoding that integer in the appropriate number of charcters
        code = ''
        for i in range(self.d):
            a = num//(64**(self.d-i-1))
            num = num - a*(64**(self.d-i-1))
            code+=self.list[a]
        
        return code[::-1]















#A method that takes as input a regina triangulation of a 3 manifold and outputs its oriented isomorphism signature.

def oIsoSig(Tri, basis = None):                         #Tri = regina.engine.PacketOfTriangulation3 the input triangulation whose oriented iso sig we are determining
    if type(Tri) is not regina.engine.Triangulation3 and type(Tri) is not regina.engine.PacketOfTriangulation3: # NRH 5/13/22

#        raise TypeError('The argument passed to the function is not of type regina.engine.PacketOfTriangulation3')
        raise TypeError('The argument passed to the function is not of type regina.engine.Triangulation3 or regina.engine.PacketOfTriangulation3')

    elif not Tri.isOrientable():
        raise ValueError('The triangulation passed to the function is not orientable. Please enter an orientable triangulation.')
    

    # This is a list of permutations of S_4 in lexicographical order. It is different from the order of regina.engine.Perm4.fromPermCode(i)
    permList = [regina.engine.Perm4(0,1,2,3), regina.engine.Perm4(0,1,3,2), regina.engine.Perm4(0,2,1,3), regina.engine.Perm4(0,2,3,1), regina.engine.Perm4(0,3,1,2), regina.engine.Perm4(0,3,2,1), 
                regina.engine.Perm4(1,0,2,3), regina.engine.Perm4(1,0,3,2), regina.engine.Perm4(1,2,0,3), regina.engine.Perm4(1,2,3,0), regina.engine.Perm4(1,3,0,2), regina.engine.Perm4(1,3,2,0), 
                regina.engine.Perm4(2,0,1,3), regina.engine.Perm4(2,0,3,1), regina.engine.Perm4(2,1,0,3), regina.engine.Perm4(2,1,3,0), regina.engine.Perm4(2,3,0,1), regina.engine.Perm4(2,3,1,0),
                regina.engine.Perm4(3,0,1,2), regina.engine.Perm4(3,0,2,1), regina.engine.Perm4(3,1,0,2), regina.engine.Perm4(3,1,2,0), regina.engine.Perm4(3,2,0,1), regina.engine.Perm4(3,2,1,0)]
    
    rightList = ['!', [1,2,3], [2,3,1], [3,1,2], [10,13,12], [13,12,10], [12,10,13], [20,21,23], [21,23,20], [23,20,21], [30,32,31], [32,31,30], [31,30,32]]


    leftList = ['?', [1,3,2], [3,2,1], [2,1,3], [10,12,13], [12,13,10], [13,10,12], [20,23,21], [23,21,20], [21,20,23], [30,31,32], [31,32,30], [32,30,31]]

    d = math.floor(math.log(Tri.size(),64)) + 1  
    piD = Pi(d)
    pi1 = Pi(1)

    oIsoSig = 'z'     #oriented isomorphism signature This is just a place holder for comparison at the moment
    
    if basis == None or basis == '$':
        firstChar = '$'
    elif basis in rightList:
        firstChar = '!'
    elif basis in leftList:
        firstChar = '?'
    else:
        raise ValueError('The basis passed to the function is not an ordered basis consisting of three edges in a single tetrahedron considered as vectors based at a common vertex')  
        #common vertex here means within the tetrahedron not in the triangulation as a whole.


    for n in range(Tri.size()):
        for s in regina.Perm4.S4:
            pList = [s]                             #List of permutations from tets in Tri to tets in T (see constuction below)
            tList = [Tri.simplex(n)]                #List of tets in Tri in order corresponding to their new order in T
            T = regina.Triangulation3()             #The canonical labelling triangulation
            T.newSimplex()
            if Tri.size() < 63:
                protoSig = firstChar + piD.encode(Tri.size())
            else:
                protoSig = firstChar + pi1.encode(63) + pi1.encode(d) + piD.encode(Tri.size())
                
            dSeq = []                       #destination sequence
            tSeq = []                       #type sequence
            pSeq = []                       #permutation sequence
            flag = False

            for i in range(Tri.size()):             #Build up T via oriented canonical labelling rules as well as constructing dSeq, tSeq, and pSeq
                for j in range(4):
                    if T.simplex(i).adjacentSimplex(j) is None:                 #check to make sure no faces are being glued twice
                        gluing = tList[i].adjacentGluing(pList[i].pre(j))
                        gTet = tList[i].adjacentSimplex(pList[i].pre(j))
                        if gTet is None:                                    #protocol for faces of type 0
                            tSeq.append(0)
                            if len(tSeq) % 3 == 0:
                                protoSig += pi1.encode(tSeq[-3] + 4*tSeq[-2] + 16*tSeq[-1])
                                if protoSig > oIsoSig:
                                    flag = True
                                    break
                        elif gTet not in tList:                             #protocol for faces of type 1
                            tList.append(gTet)
                            T.newSimplex()
                            # The following if statements glue up faces to new tetrahedra via an orientation reversing face pairing.
                            # This ensures that the resulting canonically labeled triangulation T is oriented.
                            if j == 0 or j == 1:                        
                                T.simplex(i).join(j, T.simplex(T.size()-1), regina.engine.Perm4.fromPermCode2(1))
                                pList.append(regina.engine.Perm4.fromPermCode2(1)*pList[i]*gluing.inverse())
                            elif j == 2:
                                T.simplex(i).join(j, T.simplex(T.size()-1), regina.engine.Perm4.fromPermCode2(5))
                                pList.append(regina.engine.Perm4.fromPermCode2(5)*pList[i]*gluing.inverse())
                            else:
                                T.simplex(i).join(j, T.simplex(T.size()-1), regina.engine.Perm4.fromPermCode2(3))
                                pList.append(regina.engine.Perm4.fromPermCode2(3)*pList[i]*gluing.inverse())
                            tSeq.append(1)
                            if len(tSeq) % 3 == 0:
                                protoSig += pi1.encode(tSeq[-3] + 4*tSeq[-2] + 16*tSeq[-1])
                                if protoSig > oIsoSig:
                                    flag = True
                                    break
                        else:                                               #protocol for faces of type 2
                            dest = tList.index(gTet)
                            T.simplex(i).join(j, T.simplex(dest), pList[dest]*gluing*pList[i].inverse())
                            dSeq.append(dest)
                            tSeq.append(2)
                            pSeq.append(permList.index(pList[dest]*gluing*pList[i].inverse()))
                            if len(tSeq) % 3 == 0:
                                protoSig += pi1.encode(tSeq[-3] + 4*tSeq[-2] + 16*tSeq[-1])
                                if protoSig > oIsoSig:
                                    flag = True
                                    break
                if flag:
                    break    
            if len(tSeq) % 3 == 1:          #if statements that append the necessesary number of 0s to the type sequence
                tSeq.append(0)
                tSeq.append(0)
                protoSig += pi1.encode(tSeq[-3]+4*tSeq[-2]+16*tSeq[-1])
            elif len(tSeq) % 3 == 2:
                tSeq.append(0)
                protoSig += pi1.encode(tSeq[-3]+4*tSeq[-2]+16*tSeq[-1])
            
            if protoSig > oIsoSig:
                continue


            #construct the proto-signature by encoding the destination sequence and then permtuation sequence
            for dd in dSeq:
                protoSig+=piD.encode(dd)
                if protoSig > oIsoSig:
                    flag = True
                    break
            if flag:
                continue

            for p in pSeq:
                protoSig+=pi1.encode(p)
                if protoSig > oIsoSig:
                    flag = True
                    break
            if flag:
                continue


            oIsoSig = protoSig
            
    

    return oIsoSig







#A method that given an orientable triangulation = Tri returns a list of 2 lists. The first list lists tetrahedrons of Tri in
# the order corresponding to their increasing relablelling in the oriented canonical labelling that gives rise to the oriented 
# isomorphism signature. The second list lists permutations of 4 elements that correspond with the relabelling of tetrahedra
# in the first list in order to achieve the oriented isomorphism signature. If printed == True then this method returns the same 
# as before but it will also print out the details of the relabelling.


def oIsoSigDetail(Tri, printed = False):            # if printed == True this function will print the mapping from Tri to its oIsoSig labelling
    if type(Tri) is not regina.engine.Triangulation3 and type(Tri) is not regina.engine.PacketOfTriangulation3: # NRH 5/13/22

#        raise TypeError('The argument passed to the function is not of type regina.engine.PacketOfTriangulation3')
        raise TypeError('The argument passed to the function is not of type regina.engine.Triangulation3 or regina.engine.PacketOfTriangulation3')

    elif not Tri.isOrientable():
        raise ValueError('The triangulation passed to the function is not orientable. Please enter an orientable triangulation.')
    

    # This is a list of permutations of S_4 in lexicographical order. It is different from the order of regina.engine.Perm4.fromPermCode(i)
    permList = [regina.engine.Perm4(0,1,2,3), regina.engine.Perm4(0,1,3,2), regina.engine.Perm4(0,2,1,3), regina.engine.Perm4(0,2,3,1), regina.engine.Perm4(0,3,1,2), regina.engine.Perm4(0,3,2,1), 
                regina.engine.Perm4(1,0,2,3), regina.engine.Perm4(1,0,3,2), regina.engine.Perm4(1,2,0,3), regina.engine.Perm4(1,2,3,0), regina.engine.Perm4(1,3,0,2), regina.engine.Perm4(1,3,2,0), 
                regina.engine.Perm4(2,0,1,3), regina.engine.Perm4(2,0,3,1), regina.engine.Perm4(2,1,0,3), regina.engine.Perm4(2,1,3,0), regina.engine.Perm4(2,3,0,1), regina.engine.Perm4(2,3,1,0),
                regina.engine.Perm4(3,0,1,2), regina.engine.Perm4(3,0,2,1), regina.engine.Perm4(3,1,0,2), regina.engine.Perm4(3,1,2,0), regina.engine.Perm4(3,2,0,1), regina.engine.Perm4(3,2,1,0)]
    
    
    d = math.floor(math.log(Tri.size(),64)) + 1  
    piD = Pi(d)
    pi1 = Pi(1)

    oIsoSig = 'z'     #oriented isomorphism signature This is just a place holder for comparison at the moment
    
    

    for n in range(Tri.size()):
        for s in regina.Perm4.S4:
            pList = [s]                             #List of permutations from tets in Tri to tets in T (see constuction below)
            tList = [Tri.simplex(n)]                #List of tets in Tri in order corresponding to their new order in T
            T = regina.Triangulation3()             #The canonical labelling triangulation
            T.newSimplex()
            if Tri.size() < 63:
                protoSig = '$' + piD.encode(Tri.size())
            else:
                protoSig = '$' + pi1.encode(63) + pi1.encode(d) + piD.encode(Tri.size())
                
            dSeq = []                       #destination sequence
            tSeq = []                       #type sequence
            pSeq = []                       #permutation sequence
            flag = False

            for i in range(Tri.size()):             #Build up T via oriented canonical labelling rules as well as constructing dSeq, tSeq, and pSeq
                for j in range(4):
                    if T.simplex(i).adjacentSimplex(j) is None:                 #check to make sure no faces are being glued twice
                        gluing = tList[i].adjacentGluing(pList[i].pre(j))
                        gTet = tList[i].adjacentSimplex(pList[i].pre(j))
                        if gTet is None:                                    #protocol for faces of type 0
                            tSeq.append(0)
                            if len(tSeq) % 3 == 0:
                                protoSig += pi1.encode(tSeq[-3] + 4*tSeq[-2] + 16*tSeq[-1])
                                if protoSig > oIsoSig:
                                    flag = True
                                    break
                        elif gTet not in tList:                             #protocol for faces of type 1
                            tList.append(gTet)
                            T.newSimplex()
                            # The following if statements glue up faces to new tetrahedra via an orientation reversing face pairing.
                            # This ensures that the resulting canonically labeled triangulation T is oriented.
                            if j == 0 or j == 1:                        
                                T.simplex(i).join(j, T.simplex(T.size()-1), regina.engine.Perm4.fromPermCode2(1))
                                pList.append(regina.engine.Perm4.fromPermCode2(1)*pList[i]*gluing.inverse())
                            elif j == 2:
                                T.simplex(i).join(j, T.simplex(T.size()-1), regina.engine.Perm4.fromPermCode2(5))
                                pList.append(regina.engine.Perm4.fromPermCode2(5)*pList[i]*gluing.inverse())
                            else:
                                T.simplex(i).join(j, T.simplex(T.size()-1), regina.engine.Perm4.fromPermCode2(3))
                                pList.append(regina.engine.Perm4.fromPermCode2(3)*pList[i]*gluing.inverse())
                            tSeq.append(1)
                            if len(tSeq) % 3 == 0:
                                protoSig += pi1.encode(tSeq[-3] + 4*tSeq[-2] + 16*tSeq[-1])
                                if protoSig > oIsoSig:
                                    flag = True
                                    break
                        else:                                               #protocol for faces of type 2
                            dest = tList.index(gTet)
                            T.simplex(i).join(j, T.simplex(dest), pList[dest]*gluing*pList[i].inverse())
                            dSeq.append(dest)
                            tSeq.append(2)
                            pSeq.append(permList.index(pList[dest]*gluing*pList[i].inverse()))
                            if len(tSeq) % 3 == 0:
                                protoSig += pi1.encode(tSeq[-3] + 4*tSeq[-2] + 16*tSeq[-1])
                                if protoSig > oIsoSig:
                                    flag = True
                                    break
                if flag:
                    break    
            if len(tSeq) % 3 == 1:          #if statements that append the necessesary number of 0s to the type sequence
                tSeq.append(0)
                tSeq.append(0)
                protoSig += pi1.encode(tSeq[-3]+4*tSeq[-2]+16*tSeq[-1])
            elif len(tSeq) % 3 == 2:
                tSeq.append(0)
                protoSig += pi1.encode(tSeq[-3]+4*tSeq[-2]+16*tSeq[-1])
            
            if protoSig > oIsoSig:
                continue


            #construct the proto-signature by encoding the destination sequence and then permtuation sequence
            for dd in dSeq:
                protoSig+=piD.encode(dd)
                if protoSig > oIsoSig:
                    flag = True
                    break
            if flag:
                continue

            for p in pSeq:
                protoSig+=pi1.encode(p)
                if protoSig > oIsoSig:
                    flag = True
                    break
            if flag:
                continue


            oIsoSig = protoSig
            map = [tList, pList]            #list of lists to be returned by the function
            
    if printed:                         #print the mapping from Tri to the oriented canonical labelling that gives rise to its oIsoSig
        for i in range(len(map[0])):
            print(str(map[0][i].index()) + '(0123) --> ' + str(i) + '(' + map[1][i].str() + ')')

    return map





#A method that takes as input an oriented isomorphism signature and constructs a triangulation that is a representative
# of the isomorphism class corresponding to the oriented isomorphism signature. In particular this triangulation is the
# oriented canonical labelling that gives rise tothe oriented isomorphism signature.

def fromOIsoSig(oIsoSig):
    if type(oIsoSig) is not str:
        raise TypeError('The argument passed to the function is not of type string')
    elif oIsoSig[0] != '!' and oIsoSig[0] != '?' and oIsoSig[0] != '$':
        raise ValueError('The argument passed to the function is not an oriented isomorphism signature')

    permList = [regina.engine.Perm4(0,1,2,3), regina.engine.Perm4(0,1,3,2), regina.engine.Perm4(0,2,1,3), regina.engine.Perm4(0,2,3,1), regina.engine.Perm4(0,3,1,2), regina.engine.Perm4(0,3,2,1), #List of permutations in order for encoding the permutation sequence
                regina.engine.Perm4(1,0,2,3), regina.engine.Perm4(1,0,3,2), regina.engine.Perm4(1,2,0,3), regina.engine.Perm4(1,2,3,0), regina.engine.Perm4(1,3,0,2), regina.engine.Perm4(1,3,2,0), 
                regina.engine.Perm4(2,0,1,3), regina.engine.Perm4(2,0,3,1), regina.engine.Perm4(2,1,0,3), regina.engine.Perm4(2,1,3,0), regina.engine.Perm4(2,3,0,1), regina.engine.Perm4(2,3,1,0),
                regina.engine.Perm4(3,0,1,2), regina.engine.Perm4(3,0,2,1), regina.engine.Perm4(3,1,0,2), regina.engine.Perm4(3,1,2,0), regina.engine.Perm4(3,2,0,1), regina.engine.Perm4(3,2,1,0)]

    tSeq = []                   # the type sequence
    dSeq = []                   # the destination sequence
    pSeq = []                   # the permutation sequence
    pos = 1                     # position of the character in oIsoSig that we are looking at (skips first character)

    if '#' in oIsoSig:
        oIsoSigEnd = oIsoSig.index('#')-1
    else:
        oIsoSigEnd = len(oIsoSig)

    T = regina.Triangulation3()
    if oIsoSig[pos] == 'a':     # special case for empty triangulation
        return T
    else:
        if oIsoSig[pos] != '-':                                 #parse oIsoSig preamble for number of tets < 63
            d = 1
            for i in range(Pi.decode(oIsoSig[pos])):      
                T.newSimplex()
            pos += 1
        else:                                                   #parse oIsoSig preamble for number of tets >= 63
            pos += 1
            d = Pi.list.index(oIsoSig[pos])
            pos += 1
            for i in range(Pi.decode(oIsoSig[pos:pos+d:1])):      
                T.newSimplex()
            pos += d

    
    count = 0                             # Count the number of times 2 appears in the type sequence
    for i in range(oIsoSigEnd):          # Construct type sequence for connected, closed triangulation
        flag = False       
        for j in range(3):
            for k in range(3):
                for l in range(3):
                    if j + 4*k + 16*l == Pi.decode(oIsoSig[pos]):
                        tSeq.append(j)
                        tSeq.append(k)
                        tSeq.append(l)
                        flag = True
                        if j == 2:
                            count += 1
                        if k == 2:
                            count += 1
                        if l == 2:
                            count += 1
                        pos += 1
                        break
                if flag:
                    break
            if flag:
                break
        
        if pos == oIsoSigEnd-d*count-count:
            break

    for i in range(count):                              # construct the destination and permutation sequences
        dSeq.append(Pi.decode(oIsoSig[pos:pos+d:1]))
        pos += d
    for i in range(count):
        pSeq.append(Pi.decode(oIsoSig[pos+i]))

    tPos = 0                        # position of term in type sequence
    dpPos = 0                       # position of term in dest/perm sequence
    tet = 0
    for i in range(T.size()):       # construct the oriented canonical label associated with oIsoSig
        for j in range(4):
            if T.simplex(i).adjacentSimplex(j) is None:         #check to make sure no faces are being doubly glued
                if tSeq[tPos] == 0:                             #if face is type 0 then it doesn't glue to anything
                    pass
                elif tSeq[tPos] == 1:                           #if face is type 1 then gluing depends on face number based off orientation reversing face pairings
                    tet += 1
                    if j == 0 or j == 1:
                        T.simplex(i).join(j, T.simplex(tet), regina.Perm4.fromPermCode2(1))
                    elif j == 2:
                        T.simplex(i).join(j, T.simplex(tet), regina.Perm4.fromPermCode2(5))
                    else:
                        T.simplex(i).join(j, T.simplex(tet), regina.Perm4.fromPermCode2(3))
                elif tSeq[tPos] == 2:                                                       # if face is type 2 get appropriate gluing info from destination and permutation sequences                                    
                    T.simplex(i).join(j, T.simplex(dSeq[dpPos]), permList[pSeq[dpPos]])
                    dpPos += 1
                tPos += 1

    return T










#A method that given a triangulation returns a list of 4 lists corresponding with the increasing lexicographical order of 
# subsimplices in the triangulation. The first list in the list is the vertices of the triangulation in order, the second 
# is a list of the edges in order, the third is a list of the faces in order, and the final is a list of the tetrahedra in order.

def orderSkeleton(T):
    vList = []                      #lexicographical list of vertices in T
    eList = []                      #lexicographical list of edges in T
    fList = []                      #lexicographical list of faces in T
    tList = []                      #lexicographical list of tetraherda in T
    for i in range(T.size()):
        tList.append(T.simplex(i))
        for j in range(4):
            if T.simplex(i).face(0,j) not in vList:
                vList.append(T.simplex(i).face(0,j))
            if T.simplex(i).face(1,j) not in eList:
                eList.append(T.simplex(i).face(1,j))
            if T.simplex(i).face(2,j) not in fList:
                fList.append(T.simplex(i).face(2,j))
            if j == 3:
                if T.simplex(i).face(1,4) not in eList:
                    eList.append(T.simplex(i).face(1,4))
                if T.simplex(i).face(1,5) not in eList:
                    eList.append(T.simplex(i).face(1,5))
    skeleton = [vList, eList, fList, tList]
    return skeleton
        




#A method that given Tri = Triangulation, tet = a simplex in Tri, subdim = 0, 1, 2, or 3 based off the dimension of the subsimplex
# where the Pachner move is to be performed, and location = the local index of the subsimplex within tet where the move is to occur
# returns a string that encodes the corresponding Pachner arc.

def encodePachner(Tri, tet, subdim, location):
    map = oIsoSigDetail(Tri,)
    if subdim == 0 or subdim == 2:                              #This section assigns the new location of the pachner move site after relabelling to
        newLocation = map[1][map[0].index(tet)][location]       # the oriented canonical labelling that gives rise to the oriented isomorphism signature.
    elif subdim == 1:
        edgeOrder = ['01', '02', '03', '12', '13', '23']
        newEdge = str(map[1][map[0].index(tet)][int(edgeOrder[location][0])]) + str(map[1][map[0].index(tet)][int(edgeOrder[location][1])])
        if newEdge in edgeOrder:
            newLocation = edgeOrder.index(newEdge)
        else:
            newLocation = edgeOrder.index(newEdge[::-1])
    else:
        newLocation = map[0].index(tet)



    if Tri.countFaces(subdim) < 63:                         #this section handles step one in the definition for encoding pachner arcs
        num = ''
        d = 1
        piD = Pi(d)
        pi1 = Pi(1)
    else:
        d = math.floor(math.log(Tri.countFaces(subdim),64)) + 1  
        piD = Pi(d)
        pi1 = Pi(1)
        num = pi1.encode(63) + pi1.encode(d)






    T = fromOIsoSig(oIsoSig(Tri))
    if subdim != 3:                                                            
        subSimplex = T.simplex(map[0].index(tet)).face(subdim, newLocation)
    else:
        subSimplex = T.simplex(newLocation)

    skeleton = orderSkeleton(T)

    moveLocation = piD.encode(skeleton[subdim].index(subSimplex))       #encode the final location of the move in the triangulation


    moveType = pi1.encode(subdim)                                       #encode the Pachner move's type as either 0,1,2,3 corresponding
                                                                        #to the dimension of the subsimplex where the move occurs
    return num + moveLocation + moveType





#A method that given an appropriate directed Pachner path encoding will construct and return the triangulation that represents
# the isomorphism class of triangulations corresponding with the final node in the directed Pachner path.


def fromPachner(pEncoding):                 # pEncoding must be of the form: oIsoSig + '#' + Pachner path encoding
    if type(pEncoding) is not str:
        raise TypeError('The argument passed to the function is not of type string')
    elif pEncoding[0] != '!' and pEncoding[0] != '?' and pEncoding[0] != '$':
        raise ValueError('The argument passed to the function does not begin with an oriented isomorphism signature')
    elif '#' not in pEncoding:
        raise ValueError('The argument passed to the function is not an encoding of a Pachner path. It should be a string with the following syntax: oriented isomorphism signature + "#" + Pachner path encoding')

    oIsoSigEnd = pEncoding.index('#')               
    pachnerPath = pEncoding[oIsoSigEnd +1:]             #separates string in pEncoding that encodes the Pachner path
    T = fromOIsoSig(pEncoding[0:oIsoSigEnd])            #creates T from oIsoSig in pEncoding
    
    pos = 0                                             #position in pachnerPath
    while pos < len(pachnerPath):
        skeleton = orderSkeleton(T)                     #order skeleton of T
        if pachnerPath[pos] != '-':                     #parse one pachner move in pEncoding suffix for number of subsimplices < 63
            moveLoc = Pi.decode(pachnerPath[pos])
            pos += 1
            moveType = Pi.decode(pachnerPath[pos])
            pos += 1
            T.pachner(skeleton[moveType][moveLoc])

        else:                                            #parse one pachner move in pEncoding suffix for number of subsimplices >= 63
            pos += 1
            d = Pi.list.index(pachnerPath[pos])
            pos += 1
            moveLoc = Pi.decode(pachnerPath[pos:pos+d:1])
            pos += d
            moveType = Pi.decode(pachnerPath[pos])
            pos += 1
            T.pachner(skeleton[moveType][moveLoc])

        T = fromOIsoSig(oIsoSig(T, pEncoding[0]))       #get T into the oriented canonical labelling configuration that gives rise to its oIsoSig
    return T












